/*
 * Memory Harness with AMBA AHB Slave interface
 *
 * (C) Copyleft 2008 Simply RISC LLP
 * AUTHORS: Cammisuli Salvatore <salcam@opencores.org>
 *          Di Paola Nunzio <ingnuccio@opencores.org>
 *
 * LICENSE:
 * This is a Free Hardware Design; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 * The above named program is distributed in the hope that it will
 * be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * DESCRIPTION:
 * Filename is a parameter, and the corresponding file content
 * must follow the rules stated in Verilog standard for the
 * $readmemh() system task.
 * For instance if you don't change the default name you just
 * have to put a text file named "memory.hex" in your simulation
 * directory with the following content inside:
 *
 *   // We start from address zero by default:
 *   1234567812345678
 *   FEDCBA9876543210
 *   // Now we jump to doubleword number 10 (i.e. address 80):
 *   @ 10
 *   02468ACE13579BDF
 */

module mem_harness (
    sys_clock_i, sys_reset_i,
    ahb_hready_o, ahb_hresp_o, ahb_hrdata_o,
    ahb_haddr_i, ahb_hwrite_i, ahb_hsize_i, ahb_hburst_i, ahb_hprot_i, ahb_htrans_i, ahb_hmastlock_i, ahb_hwdata_i
  );

  // System inputs
  input         sys_clock_i;     // System Clock
  input         sys_reset_i;     // System Reset

  // AMBA AHB Slave interface inputs
  input[63:0]   ahb_haddr_i;     // Address Bus
  input         ahb_hwrite_i;    // Write Enable
  input[2:0]    ahb_hsize_i;     // Size Transfer
  input[2:0]    ahb_hburst_i;    // Type Burst
  input[3:0]    ahb_hprot_i;     // Type Protection
  input[1:0]    ahb_htrans_i;    // Type Transfer
  input         ahb_hmastlock_i; // Locked Transfer
  input[63:0]   ahb_hwdata_i;    // Data In

  // AMBA AHB Slave interface registered outputs
  output        ahb_hready_o;    // Ready
  reg           ahb_hready_o;    // Ready
  output        ahb_hresp_o;     // Response
  reg           ahb_hresp_o;     // Response
  output[63:0]  ahb_hrdata_o;    // Data Output
  reg   [63:0]  ahb_hrdata_o;    // Data Output

  // Parameters
  parameter     addr_bits = 20;
  parameter     addr_max = (1<<addr_bits)-1;
  parameter     memfilename = "memory.hex";
  parameter     memdefaultcontent = 64'h0000000000000000;

  // Wires
  reg[63:0]     mem[addr_max:0];       // This is the memory!
  wire[63:0]    tmp_rd;                // Temporary read data
  reg[63:0]     tmp_wd;                // Temporary write data
  integer       i;                     // Index
  reg[7:0]      sel;                   // Select the Active Byte Lanes
  integer       seed;                  // Seed for random
  reg[1:0]      state;                 // Check for hready
  reg[63:0]     mem_hrdata_o;          // Save data read
  reg[63:0]     mem_hwdata_i;          // Save data write
  reg[63:0]     mem_haddr_i;           // Save address

// Initialization
`ifdef DEBUG
  initial begin
    $display("INFO: MEMH %m: Memory Harness with AMBA AHB Slave interface starting...");
    $display("INFO: MEMH %m: %0d Address Bits / %0d Doublewords / %0d Bytes Total Memory", addr_bits, addr_max+1, (addr_max+1)*8);
    for(i=0; i<=addr_max; i=i+1) mem[i] = memdefaultcontent;
    $readmemh(memfilename, mem);
    $display("INFO: MEMH %m: Memory initialization completed");
  end
`endif

  // Assignments
  assign tmp_rd = mem[ahb_haddr_i[addr_bits+2:3]];

  // Process the requests
  always @(posedge sys_clock_i) begin

    // Random assignment of hready signal
    ahb_hready_o = $dist_uniform(seed, 0, 500000); 
    
    // If hready HIGH, keep out previous data read
    if(state == 2'b1) begin

      if(ahb_hready_o==1) begin
      	// Return the data
        ahb_hrdata_o = mem_hrdata_o;

        // Return the response
      	ahb_hresp_o = `AHB_OKAY;

`ifdef DEBUG
      if(sel) $display("INFO: MEMH %m: R @ %t ns, AD=%X SEL=%X DAT=%X", $time, ahb_haddr_i, sel, ahb_hrdata_o);
`endif
 
	// Reset state
        state = 2'b0;

      end

    // If hready HIGH, keep out previous data write
    end else if(state == 2'b10) begin

      if(ahb_hready_o==1) begin

        // Store the data
        mem[mem_haddr_i[addr_bits+2:3]] = mem_hwdata_i;

        // Write a comment
`ifdef DEBUG
      if(sel) $display("INFO: MEMH %m: W @ %t ns, AD=%X SEL=%X DAT=%X", $time, ahb_haddr_i, sel, tmp_wd);
`endif

	// Reset state
        state = 2'b0;
      end

    end else begin

      mem_hrdata_o = 64'b0;
    
      // Read cycle
      if(!ahb_hwrite_i && (ahb_htrans_i==`AHB_NONSEQ || ahb_htrans_i==`AHB_SEQ)) begin

        // Prepare the data to load
        case(ahb_hsize_i)
	  3'b000: sel = (1'b1<<ahb_haddr_i[2:0]);
	  3'b001: sel = (2'b11<<(ahb_haddr_i[2:1]<<1));
          3'b010: sel = (4'b1111<<(ahb_haddr_i[2]<<2));
          3'b011: sel = 8'b11111111;
          3'b100: sel = 8'b11111111;  // Requires a 2nd access
          default: sel = 8'b00000000;
        endcase

        // If hready is HIGH, make the read immediately.
        if(ahb_hready_o==1) begin

          // Return the response
          ahb_hresp_o = `AHB_OKAY;

          // Return the data (ignore the byte select for reads)
          ahb_hrdata_o = tmp_rd;

// Write a comment
`ifdef DEBUG
      if(sel) $display("INFO: MEMH %m: R @ %t ns, AD=%X SEL=%X DAT=%X", $time, ahb_haddr_i, sel, ahb_hrdata_o);
`endif

        // Keep the data
        end else begin

          mem_hrdata_o = tmp_rd;

	  // State where keep out previous data read
          state = 2'b1;

        end

      // Write cycle
      end else if(ahb_hwrite_i && (ahb_htrans_i==`AHB_NONSEQ || ahb_htrans_i==`AHB_SEQ)) begin

        mem_hwdata_i = 64'b0;
        mem_haddr_i = 64'b0;
      
        // Prepare the data to store
        case(ahb_hsize_i)
          3'b000: sel = (1'b1<<ahb_haddr_i[2:0]);
  	  3'b001: sel = (2'b11<<(ahb_haddr_i[2:1]<<1));
          3'b010: sel = (4'b1111<<(ahb_haddr_i[2]<<2));
          3'b011: sel = 8'b11111111;
          3'b100: sel = 8'b11111111;  // Requires a 2nd access
          default: sel = 8'b00000000;
        endcase

        tmp_wd[63:56] = !sel[7] ? tmp_rd[63:56] : ahb_hwdata_i[63:56];
        tmp_wd[55:48] = !sel[6] ? tmp_rd[55:48] : ahb_hwdata_i[55:48];
        tmp_wd[47:40] = !sel[5] ? tmp_rd[47:40] : ahb_hwdata_i[47:40];
        tmp_wd[39:32] = !sel[4] ? tmp_rd[39:32] : ahb_hwdata_i[39:32];
        tmp_wd[31:24] = !sel[3] ? tmp_rd[31:24] : ahb_hwdata_i[31:24];
        tmp_wd[23:16] = !sel[2] ? tmp_rd[23:16] : ahb_hwdata_i[23:16];
        tmp_wd[15:08] = !sel[1] ? tmp_rd[15:08] : ahb_hwdata_i[15:08];
        tmp_wd[07:00] = !sel[0] ? tmp_rd[07:00] : ahb_hwdata_i[07:00];

        // If hready is HIGH, make the write immediately.
        if(ahb_hready_o==1) begin

          // Return the response
          ahb_hresp_o = `AHB_OKAY;

          // Clear the output data
          ahb_hrdata_o = 64'hZZZZZZZZZZZZZZZZ;

          // Store the data
          mem[ahb_haddr_i[addr_bits+2:3]] = tmp_wd;

// Write a comment
`ifdef DEBUG
      if(sel) $display("INFO: MEMH %m: W @ %t ns, AD=%X SEL=%X DAT=%X", $time, ahb_haddr_i, sel, tmp_wd);
`endif
        // Keep the data
        end else begin

          mem_hwdata_i = tmp_wd;
          mem_haddr_i = ahb_haddr_i;

          // Clear the output data
          ahb_hrdata_o = 64'hZZZZZZZZZZZZZZZZ;

	  // State where keep out previous data write
          state = 2'b10;

        end

      // No read/write cycle
      end else begin

        // Clear the response
        ahb_hresp_o = `AHB_OKAY;

        // Clear the output data
        ahb_hrdata_o = 64'hZZZZZZZZZZZZZZZZ;
        state = 2'b0;

      end    
    end 
  end
endmodule
