/*
 * Simply RISC S1 Testbench
 *
 * (C) Copyleft 2008 Simply RISC LLP
 * AUTHORS: Cammisuli Salvatore <salcam@opencores.org>
 *          Di Paola Nunzio <ingnuccio@opencores.org>
 *
 * LICENSE:
 * This is a Free Hardware Design; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 * The above named program is distributed in the hope that it will
 * be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * DESCRIPTION:
 * This is the testbench for the functional verification of the
 * S1 Core: it makes and instance of the S1 module to make it
 * possible to access one or more memory harnesses.
 */

`include "s1_defs.h"

module testbench ();

  /*
   * Wires
   */

  // Interrupt Requests
  wire[63:0] sys_irq;

  // AMBA AHB Master inputs / AMBA AHB Slave ouputs
  wire ahb_hready;                           // Ready
  wire ahb_hresp;			     // Response
  wire[(`AHB_DATA_WIDTH-1):0] ahb_hrdata;    // Data In

  // AMBA AHB Master outputs / AMBA AHB Slave inputs
  wire[63:0] ahb_haddr;       // Address Bus
  wire ahb_hwrite;            // Write Enable
  wire[2:0] ahb_hsize;        // Size Transfer
  wire[2:0] ahb_hburst;       // Type Burst
  wire[3:0] ahb_hprot;        // Type Protection
  wire[1:0] ahb_htrans;       // Type Transfer
  wire ahb_hmastlock;         // Locked Transer
  wire[63:0] ahb_hwdata;      // Data In

  // Separate AxiPc Input
  wire[31:0] axi_awaddr;      // Axi AW Address
  wire[31:0] axi_araddr;      // Axi AR Address
  wire[2:0]  axi_awsize;      // Axi AW Size
  wire[2:0]  axi_arsize;      // Axi AR Size
  wire[1:0]  axi_awburst;     // Axi AW Burst
  wire[1:0]  axi_arburst;     // Axi AR Burst
  wire[3:0]  axi_awcache;     // Axi AW Cache
  wire[3:0]  axi_arcache;     // Axi AR Cache
  wire[2:0]  axi_awprot;      // Axi AW Prot
  wire[2:0]  axi_arprot;      // Axi AR Prot
  wire[1:0]  axi_awlock;      // Axi AW Lock
  wire[1:0]  axi_arlock;      // Axi AR Lock
  wire       axi_awready;     // Axi AW Ready
  wire       axi_arready;     // Axi AR Ready
  wire[1:0]  axi_bresp;       // Axi B Response
  wire[1:0]  axi_rresp;       // Axi R Response
  wire       axi_awvalid;     // Axi AW Valid
  wire       axi_arvalid;     // Axi AR Valid
  wire       axi_wvalid;      // Axi W Valid
  wire       axi_rvalid;      // Axi R Valid
  wire       axi_bvalid;      // Axi B Valid
  wire[3:0]  axi_awlen;       // Axi AW Lenght
  wire[3:0]  axi_arlen;       // Axi AR Lenght

  // Separate hready wires for memory harnesses
  wire ahb_hready_RED_EXT_SEC;
  wire ahb_hready_HTRAPS;
  wire ahb_hready_TRAPS;
  wire ahb_hready_HPRIV_RESET;
  wire ahb_hready_KERNEL_text;
  wire ahb_hready_KERNEL_data;
  wire ahb_hready_MAIN;
  wire ahb_hready_RED_SEC;

  // Separate hresp wires for memory harnesses
  wire ahb_hresp_RED_EXT_SEC;
  wire ahb_hresp_HTRAPS;
  wire ahb_hresp_TRAPS;
  wire ahb_hresp_HPRIV_RESET;
  wire ahb_hresp_KERNEL_text;
  wire ahb_hresp_KERNEL_data;
  wire ahb_hresp_MAIN;
  wire ahb_hresp_RED_SEC;

  // Separate htrans wires for memory harnesses
  wire[1:0] ahb_htrans_RED_EXT_SEC;
  wire[1:0] ahb_htrans_HTRAPS;
  wire[1:0] ahb_htrans_TRAPS;
  wire[1:0] ahb_htrans_HPRIV_RESET;
  wire[1:0] ahb_htrans_KERNEL_text;
  wire[1:0] ahb_htrans_KERNEL_data;
  wire[1:0] ahb_htrans_MAIN;
  wire[1:0] ahb_htrans_RED_SEC;

  /*
   * Registers
   */

  // System signals
  reg sys_clock;
  reg sys_reset;

  /*
   * Behavior
   */

  always #1 sys_clock = ~sys_clock;
  assign sys_irq = 64'b0;

  initial begin

    // Display start message
    $display("INFO: TBENCH: Starting Simply RISC S1 Core simulation...");

    // Create VCD trace file
    $dumpfile("trace.vcd");
    $dumpvars();

    // Run the simulation
    sys_clock <= 1'b1;
    sys_reset <= 1'b1;
    #100
    sys_reset <= 1'b0;
    #14000
    $display("INFO: TBENCH: Completed Simply RISC S1 Core simulation!");
    $finish;

  end

  /*
   * Simply RISC S1 module instance
   */

  s1_top s1_top_0 (

    // System inputs
    .sys_clock_i(sys_clock),
    .sys_reset_i(sys_reset),
    .sys_irq_i(sys_irq),

    // AMBA AHB Master inputs
    .ahb_hready_i(ahb_hready),
    .ahb_hresp_i(ahb_hresp),
    .ahb_hrdata_i(ahb_hrdata),

    // AMBA AHB Master outputs
    .ahb_haddr_o(ahb_haddr),
    .ahb_hwrite_o(ahb_hwrite),
    .ahb_hsize_o(ahb_hsize),
    .ahb_hburst_o(ahb_hburst),
    .ahb_hprot_o(ahb_hprot),
    .ahb_htrans_o(ahb_htrans),
    .ahb_hmastlock_o(ahb_hmastlock),
    .ahb_hwdata_o(ahb_hwdata)

  );

  /*
   * Memory Harnesses with AMBA AHB Slave interface
   */

  // Section '.RED_EXT_SEC', segment 'text' - From PA 0000040000 to 0000047FFF and then together with
  // Section '.RED_EXT_SEC', segment 'data' - From PA 000004C000 to 000004FFFF => 16-3=13 addr_bits
  defparam mem_RED_EXT_SEC.addr_bits = 13;
  defparam mem_RED_EXT_SEC.memfilename = "mem_RED_EXT_SEC.image";
  defparam mem_RED_EXT_SEC.memdefaultcontent = 64'h0100000001000000;

  // Section '.HTRAPS', segment 'text'      - From PA 0000080000 to 0000087FFF and then together with
  // Section '.HTRAPS', segment 'data'      - From PA 000008C000 to 000008FFFF (zeroes) => 16-3=13 addr_bits
  defparam mem_HTRAPS.addr_bits = 13;
  defparam mem_HTRAPS.memfilename = "mem_HTRAPS.image";
  defparam mem_HTRAPS.memdefaultcontent = 64'h0100000001000000;

  // Section '.TRAPS', segment 'text'       - From PA 1000120000 to 1000127FFF and then together with
  // Section '.TRAPS', segment 'data'       - From PA 100012C000 to 100012FFFF (zeroes) => 16-3=13 addr_bits
  defparam mem_TRAPS.addr_bits = 13;
  defparam mem_TRAPS.memfilename = "mem_TRAPS.image";
  defparam mem_TRAPS.memdefaultcontent = 64'h0100000001000000;

  // Section '.HPRIV_RESET', segment 'text' - From PA 1000144000 to 1000144FFF => 12-3=9 addr_bits
  defparam mem_HPRIV_RESET.addr_bits = 9;
  defparam mem_HPRIV_RESET.memfilename = "mem_HPRIV_RESET.image";
  defparam mem_HPRIV_RESET.memdefaultcontent = 64'h0100000001000000;

  // Section '.KERNEL', segment 'text'      - From PA 1101834000 to 1101834FFF => 12-3=9 addr_bits
  defparam mem_KERNEL_text.addr_bits = 9;
  defparam mem_KERNEL_text.memfilename = "mem_KERNEL_text.image";
  defparam mem_KERNEL_text.memdefaultcontent = 64'h0100000001000000;

  // Section '.KERNEL', segment 'data'      - From PA 1101C34000 to 1101C34FFF => 12-3=9 addr_bits
  defparam mem_KERNEL_data.addr_bits = 9;
  defparam mem_KERNEL_data.memfilename = "mem_KERNEL_data.image";
  defparam mem_KERNEL_data.memdefaultcontent = 64'h0000000000000000;

  // Section '.MAIN', segment 'text'        - From PA 1130000000 to 113000FFFF => 16-3=13 addr_bits
  // Section '.MAIN', segment 'data'        - From PA 1170000000 but should be empty
  // Section '.USER_HEAP', segment 'data'   - From PA 1178020000 but should be empty
  // Section '.MAIN', segment 'bss'         - From PA 1178030000 but should be empty
  defparam mem_MAIN.addr_bits = 13;
  defparam mem_MAIN.memfilename = "mem_MAIN.image";
  defparam mem_MAIN.memdefaultcontent = 64'h0100000001000000;

  // Section '.RED_SEC', segment 'text'     - From PA FFF0000000 to FFF0000FFF => 12-3=9 addr_bits
  // Section '.RED_SEC', segment 'data'     - From PA FFF0010000 but should contain only an unused word
  defparam mem_RED_SEC.addr_bits = 9;
  defparam mem_RED_SEC.memfilename = "mem_RED_SEC.image";
  defparam mem_RED_SEC.memdefaultcontent = 64'h0100000001000000;

  // Decode the address and select the proper memory bank

  assign ahb_htrans_RED_EXT_SEC = ( (ahb_haddr[39:16]==24'h000004) ? ahb_htrans : 2'b00 );

  assign ahb_htrans_HTRAPS = ( (ahb_haddr[39:16]==24'h000008) ? ahb_htrans : 2'b00 );

  assign ahb_htrans_TRAPS = ( (ahb_haddr[39:16]==24'h100012) ? ahb_htrans : 2'b00 );

  assign ahb_htrans_HPRIV_RESET = ( (ahb_haddr[39:12]==28'h1000144) ? ahb_htrans : 2'b00 );

  assign ahb_htrans_KERNEL_text = ( (ahb_haddr[39:12]==28'h1101834) ? ahb_htrans : 2'b00 );

  assign ahb_htrans_KERNEL_data = ( (ahb_haddr[39:12]==28'h1101C34) ? ahb_htrans : 2'b00 );

  assign ahb_htrans_MAIN = ( (ahb_haddr[39:16]==24'h113000) ? ahb_htrans : 2'b00 );

  assign ahb_htrans_RED_SEC = ( (ahb_haddr[39:12]==28'hFFF0000) ? ahb_htrans : 2'b00 );

  assign ahb_hready = ahb_hready_RED_EXT_SEC & ahb_hready_HTRAPS & ahb_hready_TRAPS & ahb_hready_HPRIV_RESET &
    ahb_hready_KERNEL_text & ahb_hready_KERNEL_data & ahb_hready_MAIN & ahb_hready_RED_SEC;

  assign ahb_hresp = ahb_hresp_RED_EXT_SEC | ahb_hresp_HTRAPS | ahb_hresp_TRAPS | ahb_hresp_HPRIV_RESET | 
    ahb_hresp_KERNEL_text | ahb_hresp_KERNEL_data | ahb_hresp_MAIN | ahb_hresp_RED_SEC;

  assign axi_awaddr = ( (ahb_hwrite==1) ? ahb_haddr : 31'b0);
  assign axi_araddr = ( (ahb_hwrite==0) ? ahb_haddr : 31'b0);

  assign axi_awsize = ( (ahb_hwrite==1) ? ahb_hsize : 3'b0);
  assign axi_arsize = ( (ahb_hwrite==0) ? ahb_hsize : 3'b0);

  assign axi_awburst = ( (ahb_hwrite==1 && ahb_hburst==3'b11) ? 2'b1 : 2'b0);
  assign axi_arburst = ( (ahb_hwrite==0 && ahb_hburst==3'b11) ? 2'b1 : 2'b0);

  assign axi_awcache = ( (ahb_hwrite==1) ? { ahb_hprot[3:2], 2'b0 } : 4'b0);
  assign axi_arcache = ( (ahb_hwrite==0) ? { ahb_hprot[3:2], 2'b0 } : 4'b0);

  assign axi_awprot = ( (ahb_hwrite==1) ? { ahb_hprot[0], 1'b0, ahb_hprot[1] } : 3'b0);
  assign axi_arprot = ( (ahb_hwrite==0) ? { ahb_hprot[0], 1'b0, ahb_hprot[1] } : 3'b0);

  assign axi_awlock = ( (ahb_hwrite==1) ? { ahb_hmastlock, 1'b0 } : 2'b0);
  assign axi_arlock = ( (ahb_hwrite==0) ? { ahb_hmastlock, 1'b0 } : 2'b0);

  assign axi_awready = ( (ahb_hwrite==1) ? ahb_hready : 1'b0);
  assign axi_arready = ( (ahb_hwrite==0) ? ahb_hready : 1'b0);

  assign axi_bresp = ( (ahb_hwrite==1) ? { ahb_hresp, 1'b0 } : 2'b0);
  assign axi_rresp = ( (ahb_hwrite==0) ? { ahb_hresp, 1'b0 } : 2'b0);

  assign axi_awvalid = ( (ahb_haddr != 64'b0 && ahb_hwrite==1) ? 1'b1 : 1'b0);
  assign axi_arvalid = ( (ahb_haddr != 64'b0 && ahb_hwrite==0) ? 1'b1 : 1'b0);

  assign axi_wvalid = ( (ahb_hwdata != 64'bZ) ? 1'b1 : 1'b0);
  assign axi_rvalid = ( (ahb_hrdata != 64'bZ) ? 1'b1 : 1'b0);

  assign axi_awlen = ( (ahb_hburst == 3'b11 && ahb_hwrite==1) ? 4'b100 : 4'b0);
  assign axi_arlen = ( (ahb_hburst == 3'b11 && ahb_hwrite==0) ? 4'b100 : 4'b0);

  `ifdef OVL_ASSERT_ON

    AxiPC AxiPC_0 (

      // System inputs
      .ACLK(sys_clock),
      .ARESETn(sys_reset),

      // Write Address Channel
      .AWID(4'b0),
      .AWADDR(axi_awaddr),
      .AWLEN(axi_awlen),
      .AWSIZE(axi_awsize),
      .AWBURST(axi_awburst),
      .AWCACHE(axi_awcache),
      .AWPROT(axi_awprot),
      .AWLOCK(axi_awlock),
      .AWUSER(32'b0),
      .AWVALID(axi_awvalid),
      .AWREADY(axi_awready),

      // Write Data Channel
      .WID(4'b0),
      .WDATA(ahb_hwdata),
      .WSTRB(8'b111),
      .WUSER(32'b0),
      .WLAST(1'b0),
      .WVALID(axi_wvalid),
      .WREADY(axi_wready),

      // Write Response Channel
      .BID(4'b0),
      .BRESP(axi_bresp),
      .BUSER(32'b0),
      .BVALID(axi_wvalid),
      .BREADY(1'b1),

      // Read Address Channel
      .ARID(4'b0),
      .ARADDR(axi_araddr),
      .ARLEN(axi_arlen),
      .ARSIZE(axi_arsize),
      .ARBURST(axi_arburst),
      .ARCACHE(axi_arcache),
      .ARPROT(axi_arprot),
      .ARLOCK(axi_arlock),
      .ARUSER(32'b0),
      .ARVALID(axi_arvalid),
      .ARREADY(axi_arready),

      // Read Data Channel
      .RID(4'b0),
      .RDATA(ahb_hrdata),
      .RRESP(axi_rresp),
      .RUSER(1'b0),
      .RLAST(1'b0),
      .RVALID(axi_rvalid),
      .RREADY(axi_rready)
      
    );

  `endif

  mem_harness mem_RED_EXT_SEC (

    // System inputs
    .sys_clock_i(sys_clock),
    .sys_reset_i(sys_reset),

    // AMBA AHB Slave inputs
    .ahb_haddr_i(ahb_haddr),
    .ahb_hwrite_i(ahb_hwrite),
    .ahb_hsize_i(ahb_hsize),
    .ahb_hburst_i(ahb_hburst),
    .ahb_hprot_i(ahb_hprot),
    .ahb_htrans_i(ahb_htrans_RED_EXT_SEC),
    .ahb_hmastlock_i(ahb_hmastlock),
    .ahb_hwdata_i(ahb_hwdata),

    // AMBA AHB Slave outputs
    .ahb_hready_o(ahb_hready_RED_EXT_SEC),
    .ahb_hresp_o(ahb_hresp_RED_EXT_SEC),
    .ahb_hrdata_o(ahb_hrdata)

  );

  mem_harness mem_HTRAPS (

    // System inputs
    .sys_clock_i(sys_clock),
    .sys_reset_i(sys_reset),

    // AMBA AHB Slave inputs
    .ahb_haddr_i(ahb_haddr),
    .ahb_hwrite_i(ahb_hwrite),
    .ahb_hsize_i(ahb_hsize),
    .ahb_hburst_i(ahb_hburst),
    .ahb_hprot_i(ahb_hprot),
    .ahb_htrans_i(ahb_htrans_HTRAPS),
    .ahb_hmastlock_i(ahb_hmastlock),
    .ahb_hwdata_i(ahb_hwdata),

    // AMBA AHB Slave outputs
    .ahb_hready_o(ahb_hready_HTRAPS),
    .ahb_hresp_o(ahb_hresp_HTRAPS),
    .ahb_hrdata_o(ahb_hrdata)

  );

  mem_harness mem_TRAPS (

    // System inputs
    .sys_clock_i(sys_clock),
    .sys_reset_i(sys_reset),

    // AMBA AHB Slave inputs
    .ahb_haddr_i(ahb_haddr),
    .ahb_hwrite_i(ahb_hwrite),
    .ahb_hsize_i(ahb_hsize),
    .ahb_hburst_i(ahb_hburst),
    .ahb_hprot_i(ahb_hprot),
    .ahb_htrans_i(ahb_htrans_TRAPS),
    .ahb_hmastlock_i(ahb_hmastlock),
    .ahb_hwdata_i(ahb_hwdata),

    // AMBA AHB Slave outputs
    .ahb_hready_o(ahb_hready_TRAPS),
    .ahb_hresp_o(ahb_hresp_TRAPS),
    .ahb_hrdata_o(ahb_hrdata)

  );

  mem_harness mem_HPRIV_RESET (

    // System inputs
    .sys_clock_i(sys_clock),
    .sys_reset_i(sys_reset),

    // AMBA AHB Slave inputs
    .ahb_haddr_i(ahb_haddr),
    .ahb_hwrite_i(ahb_hwrite),
    .ahb_hsize_i(ahb_hsize),
    .ahb_hburst_i(ahb_hburst),
    .ahb_hprot_i(ahb_hprot),
    .ahb_htrans_i(ahb_htrans_HPRIV_RESET),
    .ahb_hmastlock_i(ahb_hmastlock),
    .ahb_hwdata_i(ahb_hwdata),

    // AMBA AHB Slave outputs
    .ahb_hready_o(ahb_hready_HPRIV_RESET),
    .ahb_hresp_o(ahb_hresp_HPRIV_RESET),
    .ahb_hrdata_o(ahb_hrdata)

  );

  mem_harness mem_KERNEL_text (

    // System inputs
    .sys_clock_i(sys_clock),
    .sys_reset_i(sys_reset),

    // AMBA AHB Slave inputs
    .ahb_haddr_i(ahb_haddr),
    .ahb_hwrite_i(ahb_hwrite),
    .ahb_hsize_i(ahb_hsize),
    .ahb_hburst_i(ahb_hburst),
    .ahb_hprot_i(ahb_hprot),
    .ahb_htrans_i(ahb_htrans_KERNEL_text),
    .ahb_hmastlock_i(ahb_hmastlock),
    .ahb_hwdata_i(ahb_hwdata),

    // AMBA AHB Slave outputs
    .ahb_hready_o(ahb_hready_KERNEL_text),
    .ahb_hresp_o(ahb_hresp_KERNEL_text),
    .ahb_hrdata_o(ahb_hrdata)

  );

  mem_harness mem_KERNEL_data (

    // System inputs
    .sys_clock_i(sys_clock),
    .sys_reset_i(sys_reset),

    // AMBA AHB Slave inputs
    .ahb_haddr_i(ahb_haddr),
    .ahb_hwrite_i(ahb_hwrite),
    .ahb_hsize_i(ahb_hsize),
    .ahb_hburst_i(ahb_hburst),
    .ahb_hprot_i(ahb_hprot),
    .ahb_htrans_i(ahb_htrans_KERNEL_data),
    .ahb_hmastlock_i(ahb_hmastlock),
    .ahb_hwdata_i(ahb_hwdata),

    // AMBA AHB Slave outputs
    .ahb_hready_o(ahb_hready_KERNEL_data),
    .ahb_hresp_o(ahb_hresp_KERNEL_data),
    .ahb_hrdata_o(ahb_hrdata)

  );

  mem_harness mem_MAIN (

    // System inputs
    .sys_clock_i(sys_clock),
    .sys_reset_i(sys_reset),

    // AMBA AHB Slave inputs
    .ahb_haddr_i(ahb_haddr),
    .ahb_hwrite_i(ahb_hwrite),
    .ahb_hsize_i(ahb_hsize),
    .ahb_hburst_i(ahb_hburst),
    .ahb_hprot_i(ahb_hprot),
    .ahb_htrans_i(ahb_htrans_MAIN),
    .ahb_hmastlock_i(ahb_hmastlock),
    .ahb_hwdata_i(ahb_hwdata),

    // AMBA AHB Slave outputs
    .ahb_hready_o(ahb_hready_MAIN),
    .ahb_hresp_o(ahb_hresp_MAIN),
    .ahb_hrdata_o(ahb_hrdata)

  );

  mem_harness mem_RED_SEC (

    // System inputs
    .sys_clock_i(sys_clock),
    .sys_reset_i(sys_reset),

    // AMBA AHB Slave inputs
    .ahb_haddr_i(ahb_haddr),
    .ahb_hwrite_i(ahb_hwrite),
    .ahb_hsize_i(ahb_hsize),
    .ahb_hburst_i(ahb_hburst),
    .ahb_hprot_i(ahb_hprot),
    .ahb_htrans_i(ahb_htrans_RED_SEC),
    .ahb_hmastlock_i(ahb_hmastlock),
    .ahb_hwdata_i(ahb_hwdata),

    // AMBA AHB Slave outputs
    .ahb_hready_o(ahb_hready_RED_SEC),
    .ahb_hresp_o(ahb_hresp_RED_SEC),
    .ahb_hrdata_o(ahb_hrdata)

  );

endmodule
