/*
 * Bridge from SPARC Core to AMBA AHB Master
 *
 * (C) 2008 Simply RISC LLP
 * AUTHORS: Cammisuli Salvatore <salcam@opencores.org>
 *          Di Paola Nunzio <ingnuccio@opencores.org>
 *
 * LICENSE:
 * This is a Free Hardware Design; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 * The above named program is distributed in the hope that it will
 * be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * DESCRIPTION:
 * This block implements a bridge from one SPARC Core of the
 * OpenSPARC T1 to a master interface that makes use of the
 * AMBA AHB interconnect protocol.
 * For informations about Sun Microsystems' OpenSPARC T1
 * refer to the web site http://www.opensparc.net
 * For informations about OpenCores' AMBA AHB interconnect
 * please refer to the web site http://www.opencores.org
 */

`include "s1_defs.h"

module spc2ahb (
    sys_clock_i, sys_reset_i, sys_interrupt_source_i,
    spc_req_i, spc_atom_i, spc_packetout_i,
    spc_grant_o, spc_ready_o, spc_packetin_o, spc_stallreq_o,
    ahb_hready_i, ahb_hresp_i, ahb_hrdata_i,
    ahb_haddr_o, ahb_hwrite_o, ahb_hsize_o, ahb_hburst_o, ahb_hprot_o, ahb_htrans_o, ahb_hmastlock_o, ahb_hwdata_o
  );

  /*
   * Inputs
   */

  // System inputs
  input sys_clock_i;                            // System Clock
  input sys_reset_i;                            // System Reset
  input[5:0] sys_interrupt_source_i;            // Encoded Interrupt Source

  // SPARC-side inputs connected to the PCX (Processor-to-Cache Xbar) outputs of the SPARC Core
  input[4:0] spc_req_i;                         // Request
  input spc_atom_i;                             // Atomic Request
  input[(`PCX_WIDTH-1):0] spc_packetout_i;      // Outgoing Packet

  // AMBA AHB Master interface inputs
  input ahb_hready_i;                           // Ready
  input ahb_hresp_i;				// Response
  input[(`AHB_DATA_WIDTH-1):0] ahb_hrdata_i;    // Data In

  /*
   * Registered Outputs
   */

  // SPARC-side outputs connected to the CPX (Cache-to-Processor Xbar) inputs of the SPARC Core
  output[4:0] spc_grant_o;                      // Grant
  reg[4:0] spc_grant_o;                         // Grant
  output spc_ready_o;                           // Ready
  reg spc_ready_o;                              // Ready
  output[`CPX_WIDTH-1:0] spc_packetin_o;        // Incoming Packet
  reg[`CPX_WIDTH-1:0] spc_packetin_o;           // Incoming Packet
  output spc_stallreq_o;                        // Stall Request
  reg spc_stallreq_o;                           // Stall Request

  // AMBA AHB Master interface outputs
  output[`AHB_ADDR_WIDTH-1:0] ahb_haddr_o;      // Address Bus
  reg[`AHB_ADDR_WIDTH-1:0] ahb_haddr_o;         // Address Bus
  output ahb_hwrite_o;                          // Write Enable
  reg ahb_hwrite_o;                             // Write Enable
  output[`AHB_SIZE_WIDTH-1:0] ahb_hsize_o;      // Size Transfer
  reg[`AHB_SIZE_WIDTH-1:0] ahb_hsize_o;         // Size Transfer
  output[`AHB_BURST_WIDTH-1:0] ahb_hburst_o;    // Type Burst
  reg[`AHB_BURST_WIDTH-1:0] ahb_hburst_o;       // Type Burst
  output[`AHB_PROT_WIDTH-1:0] ahb_hprot_o;      // Type Protection
  reg[`AHB_PROT_WIDTH-1:0] ahb_hprot_o;         // Type Protection
  output[`AHB_TRANS_WIDTH-1:0] ahb_htrans_o;    // Type Transfer
  reg[`AHB_TRANS_WIDTH-1:0] ahb_htrans_o;       // Type Transfer
  output ahb_hmastlock_o;                       // Locked Transfer
  reg ahb_hmastlock_o;                          // Locked Transfer
  output[`AHB_DATA_WIDTH-1:0] ahb_hwdata_o;     // Data Out
  reg[`AHB_DATA_WIDTH-1:0] ahb_hwdata_o;        // Data Out

  /*
   * Registers
   */

  // Registers to latch requests from SPARC Core to AMBA AHB Master
  reg[1:0] state;						       // State of FMS
  reg req_enable;						       // Request phase
  reg packet_arrived;						       // Packet arrived
  reg address_data_phase; 					       // Address & data phase
  reg[2:0] waited_phase;					       // Waited Phase
  reg[2:0] burst_phase;						       // Burst phase
  reg ahb_working;						       // Bridge processing the request
  reg[(`PCX_WIDTH-1):0] data_saved;  				       // Save the data, while bridge working
  reg read_data_saved;						       // Read the saved data
  reg waited_burst;						        // Check the burst in the waited phase

  reg[4:0] spc2ahb_region;                                             // Target region number (one-hot encoded)
  reg spc2ahb_atomic;                                                  // Request is Atomic
  reg[(`PCX_WIDTH-1):0] spc2ahb_packet;                                // Latched Packet
  reg[(`PCX_DA_HI-`PCX_DA_LO):0] spc2ahb_data;                         // Store Data

  // AMBA AHB Master to SPARC Core info used to encode the return packet
  reg ahb2spc_valid;                                                   // Valid
  reg[(`CPX_RQ_HI-`CPX_RQ_LO):0] ahb2spc_type;                         // Request type
  reg ahb2spc_miss;                                                    // L2 Miss
  reg[(`CPX_ERR_HI-`CPX_ERR_LO-1):0] ahb2spc_error;                    // Error
  reg ahb2spc_nc;                                                      // Non-Cacheable
  reg[(`CPX_TH_HI-`CPX_TH_LO):0] ahb2spc_thread;                       // Thread
  reg ahb2spc_way_valid;                                               // L2 Way Valid
  reg[(`CPX_WY_HI-`CPX_WY_LO):0] ahb2spc_way;                          // Replaced L2 Way
  reg ahb2spc_boot_fetch;                                              // Fetch for Boot
  reg ahb2spc_atomic;                                                  // Atomic LD/ST or 2nd IFill packet
  reg ahb2spc_pfl;                                                     // PFL
  reg[(`CPX_DA_HI-`CPX_DA_LO):0] ahb2spc_data;                         // Load Data
  reg[6:0] ahb2spc_interrupt_source;                                   // Encoded Interrupt Source
  reg ahb2spc_interrupt_new;                                           // New Interrupt Pending

  reg[2:0] state_prev;                                                 // Check the previous state
  reg burst_prev;                                                      // Check if the previous state is a burst
  reg packet_ready;                                                    // Packet Ready
  integer seed; 	                  			       // Seed for random
   
  /*
   * Wires
   */

  // Decoded SPARC Core to AMBA AHB Master info
  wire spc2ahb_req;                                                     // Request
  wire spc2ahb_valid;                                                   // Valid
  wire[(`PCX_RQ_HI-`PCX_RQ_LO):0] spc2ahb_type;                         // Request type
  wire spc2ahb_nc;                                                      // Non-Cacheable
  wire[(`PCX_CP_HI-`PCX_CP_LO):0] spc2ahb_cpu_id;                       // CPU ID
  wire[(`PCX_TH_HI-`PCX_TH_LO):0] spc2ahb_thread;                       // Thread
  wire spc2ahb_invalidate;                                              // Invalidate all
  wire[(`PCX_WY_HI-`PCX_WY_LO):0] spc2ahb_way;                          // Replaced L1 Way
  wire[(`PCX_SZ_HI-`PCX_SZ_LO):0] spc2ahb_size;                         // Load/Store size
  wire[(`PCX_AD_HI-`PCX_AD_LO):0] spc2ahb_addr;                         // Address

  // Return packets assembled with various fields
  wire[`CPX_WIDTH-1:0] ahb2spc_packet;                                  // Incoming Packet

  /*
   * Encode/decode incoming info
   *
   * Legenda: available constants for some of the PCX/CPX fields.
   *
   * spc2ahb_size (3 bits) is one of:
   * - PCX_SZ_1B
   * - PCX_SZ_2B
   * - PCX_SZ_4B
   * - PCX_SZ_8B
   * - PCX_SZ_16B (Read accesses only)
   *
   * spc2ahb_type (5 bits) is one of:
   * { LOAD_RQ, IMISS_RQ, STORE_RQ, CAS1_RQ, CAS2_RQ, SWAP_RQ, STRLOAD_RQ, STRST_RQ, STQ_RQ,
   *   INT_RQ, FWD_RQ, FWD_RPY, RSVD_RQ }
   *
   * ahb2spc_type (4 bits) is one of:
   * { LOAD_RET, INV_RET, ST_ACK, AT_ACK, INT_RET, TEST_RET, FP_RET, IFILL_RET, EVICT_REQ,
   *   ERR_RET, STRLOAD_RET, STRST_ACK, FWD_RQ_RET, FWD_RPY_RET, RSVD_RET }
   *
   */

  // Decode info arriving from the SPC side
  assign spc2ahb_req = ( spc_req_i[4] | spc_req_i[3] | spc_req_i[2] | spc_req_i[1] | spc_req_i[0] );
  assign spc2ahb_valid = spc2ahb_packet[`PCX_VLD];
  assign spc2ahb_type = spc2ahb_packet[`PCX_RQ_HI:`PCX_RQ_LO];
  assign spc2ahb_nc = spc2ahb_packet[`PCX_NC];
  assign spc2ahb_cpu_id = spc2ahb_packet[`PCX_CP_HI:`PCX_CP_LO];
  assign spc2ahb_thread = spc2ahb_packet[`PCX_TH_HI:`PCX_TH_LO];
  assign spc2ahb_invalidate = spc2ahb_packet[`PCX_INVALL];
  assign spc2ahb_way = spc2ahb_packet[`PCX_WY_HI:`PCX_WY_LO];
  assign spc2ahb_size = spc2ahb_packet[`PCX_SZ_HI:`PCX_SZ_LO];
  assign spc2ahb_addr = spc2ahb_packet[`PCX_AD_HI:`PCX_AD_LO];
  
  // Encode info going to the SPC side assembling return packets
  assign ahb2spc_packet = { ahb2spc_valid, ahb2spc_type, ahb2spc_miss, ahb2spc_error, ahb2spc_nc, ahb2spc_thread,
    ahb2spc_way_valid, ahb2spc_way, ahb2spc_boot_fetch, ahb2spc_atomic, ahb2spc_pfl, ahb2spc_data };

  /*
   * State Machine
   */

  always @(posedge sys_clock_i) begin

    // Initialization
    if(sys_reset_i==1) begin

      // Clear outputs going to SPARC Core inputs
      spc_grant_o <= 5'b00000;
      spc_ready_o <= 0;
      spc_packetin_o <= 0;
      spc_stallreq_o <= 0;

      // Clear AMBA AHB Master interface outputs
      ahb_haddr_o <= `AHB_ADDR_WIDTH'b0;
      ahb_hwrite_o <= 0;
      ahb_hsize_o <= `AHB_SIZE_WIDTH'b0;
      ahb_hburst_o <= `AHB_SINGLE;
      ahb_hprot_o <= `AHB_PROT_WIDTH'b0;
      ahb_htrans_o <= `AHB_IDLE;
      ahb_hmastlock_o <= 0;
      ahb_hwdata_o <= `AHB_DATA_WIDTH'bZ;

      // Prepare wakeup packet for SPARC Core, the resulting output is
      ahb2spc_valid <= 1;
      ahb2spc_type <= `INT_RET;
      ahb2spc_miss <= 0;
      ahb2spc_error <= 0;
      ahb2spc_nc <= 0;
      ahb2spc_thread <= 0;
      ahb2spc_way_valid <= 0;
      ahb2spc_way <= 0;
      ahb2spc_boot_fetch <= 0;
      ahb2spc_atomic <= 0;
      ahb2spc_pfl <= 0;
      ahb2spc_data <= 64'h10001;
      ahb2spc_interrupt_source <= 7'h0;
      ahb2spc_interrupt_new <= 1'b0;

      // Clear state machine
      state <= `STATE_WAKEUP;
      burst_phase <= 3'b0;
      packet_ready <= 1'b0;
      req_enable <= 1'b0;
      packet_arrived <= 1'b0;
      address_data_phase <= 1'b0;
      waited_phase <= 3'b0;     
      ahb_working <= 1'b0;
      data_saved <= `PCX_WIDTH'b0;  
      read_data_saved <= 1'b0;
      burst_prev <= 1'b0;
      waited_burst <= 1'b0;

    end else begin

      // FSM State 0: STATE_WAKEUP
      // Send to the SPARC Core the wakeup packet
      if(state==`STATE_WAKEUP) begin

        // Send wakeup packet
        spc_ready_o <= 1;
        spc_packetin_o <= ahb2spc_packet;

// Display comment
`ifdef DEBUG
        $display("INFO: SPC2AHB: SPARC Core to AMBA AHB Master bridge starting...");
        $display("INFO: SPC2AHB: Wakeup packet sent to SPARC Core");
`endif

        // Unconditional state change
        state <= `STATE_ADDRESS_DATA;

      // FSM State 1: STATE_ADDRESS_DATA
      // Wait for a request from the SPARC Core
      // Address and Data phase are implemented in AHB transfer
      // If available send an interrupt packet to the Core
      end else if(state==`STATE_ADDRESS_DATA) begin

	// Data & Address Phase
        if(packet_arrived==1'b1 && burst_phase==3'b0) begin
          
          spc_grant_o <= 5'b0;

          if(ahb_hready_i==1'b1) begin

            // Issue a request on the AMBA AHB bus
  	    ahb_hwdata_o <= spc2ahb_data;
	    ahb_haddr_o <= { spc2ahb_region, 19'b0, spc2ahb_addr[`PCX_AD_HI-`PCX_AD_LO:3], 3'b0 };
	    ahb_hsize_o <= spc2ahb_size;
	    ahb_hburst_o <= ( (spc2ahb_size==3'b111) ? `AHB_INCR4 : `AHB_SINGLE );
	    ahb_hprot_o <= { spc2ahb_nc, 3'b11 };
	    ahb_hmastlock_o <= spc2ahb_atomic;
	    spc2ahb_data <= spc2ahb_packet[`PCX_DA_HI:`PCX_DA_LO];	

            // Handle write enable and transfer type
            if(spc2ahb_type==`IMISS_RQ || spc2ahb_type==`LOAD_RQ) begin

              // For instruction miss always read memory
              ahb_hwrite_o <= 0;
              ahb_htrans_o <= `AHB_NONSEQ;

            end else if(spc2ahb_type==`STORE_RQ) begin

              // For data store use the provided data
              ahb_hwrite_o <= 1;
	      ahb_htrans_o <= `AHB_NONSEQ;

            end else begin

              ahb_hwrite_o <= 1;
	      ahb_htrans_o <= `AHB_IDLE;

            end          

// Print details of request packet
`ifdef DEBUG
        $display("INFO: SPC2AHB: Valid bit is %X", spc2ahb_valid);
        case(spc2ahb_type)
          `LOAD_RQ: $display("INFO: SPC2AHB: Request of Type LOAD_RQ");
          `IMISS_RQ: $display("INFO: SPC2AHB: Request of Type IMISS_RQ");
          `STORE_RQ: $display("INFO: SPC2AHB: Request of Type STORE_RQ");
          `CAS1_RQ: $display("INFO: SPC2AHB: Request of Type CAS1_RQ");
          `CAS2_RQ: $display("INFO: SPC2AHB: Request of Type CAS2_RQ");
          `SWAP_RQ: $display("INFO: SPC2AHB: Request of Type SWAP_RQ");
          `STRLOAD_RQ: $display("INFO: SPC2AHB: Request of Type STRLOAD_RQ");
          `STRST_RQ: $display("INFO: SPC2AHB: Request of Type STRST_RQ");
          `STQ_RQ: $display("INFO: SPC2AHB: Request of Type STQ_RQ");
          `INT_RQ: $display("INFO: SPC2AHB: Request of Type INT_RQ");
          `FWD_RQ: $display("INFO: SPC2AHB: Request of Type FWD_RQ");
          `FWD_RPY: $display("INFO: SPC2AHB: Request of Type FWD_RPY");
          `RSVD_RQ: $display("INFO: SPC2AHB: Request of Type RSVD_RQ");
          default: $display("INFO: SPC2AHB: Request of Type Unknown");
	endcase
        $display("INFO: SPC2AHB: Non-Cacheable bit is %X", spc2ahb_nc);
        $display("INFO: SPC2AHB: CPU-ID is %X", spc2ahb_cpu_id);
        $display("INFO: SPC2AHB: Thread is %X", spc2ahb_thread);
        $display("INFO: SPC2AHB: Invalidate All is %X", spc2ahb_invalidate);
        $display("INFO: SPC2AHB: Replaced L1 Way is %X", spc2ahb_way);
        case(spc2ahb_size)
          `PCX_SZ_1B: $display("INFO: SPC2AHB: Request size is 1 Byte");
          `PCX_SZ_2B: $display("INFO: SPC2AHB: Request size is 2 Bytes");
          `PCX_SZ_4B: $display("INFO: SPC2AHB: Request size is 4 Bytes");
          `PCX_SZ_8B: $display("INFO: SPC2AHB: Request size is 8 Bytes");
          `PCX_SZ_16B: $display("INFO: SPC2AHB: Request size is 16 Bytes");
          default: $display("INFO: SPC2AHB: Request size is Unknown");
        endcase
        $display("INFO: SPC2AHB: Address is %X", spc2ahb_addr);
        $display("INFO: SPC2AHB: Data is %X", spc2ahb_data);
`endif

	    packet_arrived <= 1'b0;
	    waited_phase <= 3'b1;

	  end else state <= `STATE_WAITED;

        end

        // Waited phase to synchronize processor and bridge
        if(waited_phase==3'b1)begin

	  // Clear previously modified outputs
          ahb_haddr_o <= 64'b0;
          ahb_hwrite_o <= 1'b0;
          ahb_hsize_o <= 3'b0;
	  ahb_hprot_o <= 4'b0;
          ahb_htrans_o <= `AHB_IDLE;
          ahb_hmastlock_o <= 1'b0;
          ahb_hwdata_o <= 64'bZ;

	  waited_phase <= 3'b0;
          address_data_phase <= 1'b1;
	    
        end

        // Waited phase to synchronize processor and bridge
        if(waited_phase==3'b10)begin

          // Clear previously modified outputs
          ahb_haddr_o <= 64'b0;
          ahb_hwrite_o <= 1'b0;
          ahb_hsize_o <= 3'b0;
	  ahb_hprot_o <= 4'b0;
          ahb_htrans_o <= `AHB_IDLE;
          ahb_hmastlock_o <= 1'b0;
          ahb_hwdata_o <= 64'bZ;

	  waited_phase <= 3'b0;
          burst_phase <= 3'b001;
	    
        end

        // Waited phase to synchronize processor and bridge
 	if(waited_phase==3'b11)begin

          // Clear previously modified outputs
          ahb_haddr_o <= 64'b0;
          ahb_hwrite_o <= 1'b0;
          ahb_hsize_o <= 3'b0;
	  ahb_hprot_o <= 4'b0;
          ahb_htrans_o <= `AHB_IDLE;
          ahb_hmastlock_o <= 1'b0;
          ahb_hwdata_o <= 64'bZ;

	  waited_phase <= 3'b0;
          burst_phase <= 3'b010;
          waited_burst <= 1'b1;
	    
        end

        // Waited phase to synchronize processor and bridge
        if(waited_phase==3'b100)begin

          // Clear previously modified outputs
          ahb_haddr_o <= 64'b0;
          ahb_hwrite_o <= 1'b0;
          ahb_hsize_o <= 3'b0;
	  ahb_hprot_o <= 4'b0;
          ahb_htrans_o <= `AHB_IDLE;
          ahb_hmastlock_o <= 1'b0;
          ahb_hwdata_o <= 64'bZ;

	  waited_phase <= 3'b0;
          burst_phase <= 3'b011;
          waited_burst <= 1'b0;
	    
        end

	// Return packet
	if(address_data_phase==1) begin
  
          state_prev <= 3'b1;

          // Wait until AMBA AHB access completes
          if(ahb_hresp_i==`AHB_OKAY && ahb_hready_i==1) begin
            state_prev <= 3'b0;
            address_data_phase <= 1'b0;

            // Latch the data and set up the return packet for the SPARC Core
            ahb2spc_valid <= 1;
            case(spc2ahb_type)
              `IMISS_RQ: begin
                ahb2spc_type <= `IFILL_RET; // I-Cache Miss
                ahb2spc_atomic <= 0;
              end
              `LOAD_RQ: begin
                ahb2spc_type <= `LOAD_RET;  // Load
                ahb2spc_atomic <= spc2ahb_atomic;
              end
              `STORE_RQ: begin
                ahb2spc_type <= `ST_ACK;    // Store
                ahb2spc_atomic <= spc2ahb_atomic;
              end
            endcase
            ahb2spc_miss <= 0;
            ahb2spc_error <= 0;
            ahb2spc_nc <= spc2ahb_nc;
            ahb2spc_thread <= spc2ahb_thread;
            ahb2spc_way_valid <= 0;
            ahb2spc_way <= 0;
            if(spc2ahb_region==5'b10000) ahb2spc_boot_fetch <= 1;
	    else ahb2spc_boot_fetch <= 0;
            ahb2spc_pfl <= 0;	   
            if(spc2ahb_addr[3]==0) ahb2spc_data <= { ahb_hrdata_i, 64'b0 };
            else ahb2spc_data <= { 64'b0, ahb_hrdata_i };

            // See if other 64-bit AMBA AHB accesses are required
            if(
                // Instruction miss directed to RAM expects 256 bits
                ( (spc2ahb_type==`IMISS_RQ)&&(spc2ahb_region!=5'b10000) ) ||
                // Data access of 128 bits
                ( (spc2ahb_type==`LOAD_RQ)&&(spc2ahb_size==`PCX_SZ_16B) )
              ) begin

              waited_phase <= 3'b10;

	      // Issue a second request on the AMBA AHB bus
   	      ahb_hwdata_o <= 64'bZ;
              ahb_haddr_o <= { spc2ahb_region, 19'b0, spc2ahb_addr[`PCX_AD_HI-`PCX_AD_LO:4], 4'b1000 };  // 2nd doubleword inside the same quadword
	      ahb_hwrite_o <= 0;        
	      ahb_hsize_o <= spc2ahb_size;
	      ahb_hprot_o <= { spc2ahb_nc, 3'b011 };
	      ahb_htrans_o <= `AHB_SEQ;
	      spc2ahb_data <= spc2ahb_packet[`PCX_DA_HI:`PCX_DA_LO];
            end else begin

	      burst_phase <= 3'b0;
              packet_ready <= 1'b1;

            end
  
          end else state <= `STATE_WAITED;
	  
	end
	
	// Packet Arrived Phase
	if(req_enable==1 && ahb_working==0) begin

          ahb_working <= 1'b1;
          
          if(read_data_saved==0)
            // Latch the incoming packet
            spc2ahb_packet <= spc_packetout_i;
	  else begin
            spc2ahb_packet <= data_saved;
            read_data_saved <= 1'b0;
          end

          // Grant the request to the SPARC Core
          spc_grant_o <= spc2ahb_region;     

// Print details of SPARC Core request
`ifdef DEBUG
        $display("INFO: SPC2AHB: *** NEW REQUEST FROM SPARC CORE ***");
        if(spc2ahb_region[0]==1) $display("INFO: SPC2AHB: Request to RAM Bank 0");
        else if(spc2ahb_region[1]==1) $display("INFO: SPC2AHB: Request to RAM Bank 1");
        else if(spc2ahb_region[2]==1) $display("INFO: SPC2AHB: Request to RAM Bank 2");
        else if(spc2ahb_region[3]==1) $display("INFO: SPC2AHB: Request to RAM Bank 3");
        else if(spc2ahb_region[4]==1) $display("INFO: SPC2AHB: Request targeted to I/O Block");
        else $display("INFO: SPC2AHB: Request to target region unknown");
        if(spc2ahb_atomic==1) $display("INFO: SPC2AHB: Request is ATOMIC");
        else $display("INFO: SPC2AHB: Request is not atomic");
`endif

          req_enable <= 1'b0;
	  packet_arrived <= 1'b1;     

        end
	
	// If the bridge are processing a new request, save the data
        if(req_enable==1 && ahb_working==1 && read_data_saved==0)begin
	  data_saved <= spc_packetout_i;
          read_data_saved <= 1;
        end

        // Check if there's an incoming request
        if(spc2ahb_req==1) begin

          req_enable <= 1'b1;

          // Clear previously modified outputs
          spc_ready_o <= 1'b0;
          spc_packetin_o <= 0;

          // Latch target region and atomicity
          spc2ahb_region <= spc_req_i;
          spc2ahb_atomic <= spc_atom_i;	 

        // See if the interrupt vector has changed
        end else if(sys_interrupt_source_i!=ahb2spc_interrupt_source) begin

          // Set the flag for next cycle
          ahb2spc_interrupt_new <= 1;

          // Prepare the interrupt packet for the SPARC Core
          ahb2spc_valid <= 1;
          ahb2spc_type <= `INT_RET;
          ahb2spc_miss <= 0;
          ahb2spc_error <= 0;
          ahb2spc_nc <= 0;
          ahb2spc_thread <= 0;
          ahb2spc_way_valid <= 0;
          ahb2spc_way <= 0;
          ahb2spc_boot_fetch <= 0;
          ahb2spc_atomic <= 0;
          ahb2spc_pfl <= 0;	   

          // Stall other requests from the SPARC Core
          spc_stallreq_o <= 1;

        // Next cycle see if there's an int to be forwarded to the Core
        end else if(ahb2spc_interrupt_source!=6'b000000 && ahb2spc_interrupt_new) begin

          // Clean the flag
          ahb2spc_interrupt_new <= 0;

          // Send the interrupt packet to the Core
          spc_ready_o <= 1;
          spc_packetin_o <= ahb2spc_packet;

          // Stall other requests from the SPARC Core
          spc_stallreq_o <= 1;

          // Stay in this state
          state <= `STATE_ADDRESS_DATA;

        // Nothing to do, stay idle
        end else begin

          // Clear previously modified outputs
          spc_ready_o <= 0;
	  if(burst_phase == 3'b0)
            spc_packetin_o <= 0;
          spc_stallreq_o <= 0;

        end	

        // Address phase of third burst and data phase of second burst
        if(burst_phase==3'b1) begin

          state_prev <= 3'b10;

          // Wait until AMBA AHB access completes
	  if(ahb_hresp_i==`AHB_OKAY && ahb_hready_i==1) begin

            // Return the packet to the SPARC Core
            spc_ready_o <= 1;

            state_prev <= 3'b0;

            // Latch the data and set up the return packet for the SPARC Core
	    ahb2spc_data[63:0] <= ahb_hrdata_i;
            
            // Issue a third request on the AMBA AHB bus
	    ahb_hwdata_o <= 64'bZ;
            ahb_haddr_o <= { spc2ahb_region, 19'b0, spc2ahb_addr[`PCX_AD_HI-`PCX_AD_LO:5], 5'b10000 }; //  3nd doubleword inside the same 256-bit data
            ahb_hwrite_o <= 0;        
	    ahb_hsize_o <= spc2ahb_size;
            ahb_hburst_o <= `AHB_INCR4;
	    ahb_hprot_o <= { spc2ahb_nc, 3'b011 };
	    ahb_htrans_o <= `AHB_SEQ;
	    spc2ahb_data <= spc2ahb_packet[`PCX_DA_HI:`PCX_DA_LO]; 

            // See if two return packets are required or just one
            if(spc2ahb_type==`IMISS_RQ && spc2ahb_region==5'b10000)
              packet_ready <= 1'b1;
            else begin
              waited_phase <= 3'b11;
              burst_phase <= 3'b110;
            end

          end else begin
            state <= `STATE_WAITED;   
            ahb_htrans_o <= `AHB_IDLE;
          end
        // Address phase of fourth burst and data phase of third burst
        end else if(burst_phase == 3'b10) begin

          state_prev <= 3'b11;
          burst_prev <= 1'b0;
          waited_burst <= 1'b0;
          // Wait until AMBA AHB access completes
	  if(ahb_hresp_i==`AHB_OKAY && ahb_hready_i==1) begin

            // Clear previously modified outputs
            spc_ready_o <= 0;

            // Latch the data and set up the return packet for the SPARC Core
            ahb2spc_data <= { ahb_hrdata_i, 64'b0 };
	    state_prev <= 3'b0;

            // Issue a fourth request on the AMBA AHB bus
	    ahb_hwdata_o <= 64'bZ;
            ahb_haddr_o <= { spc2ahb_region, 19'b0, spc2ahb_addr[`PCX_AD_HI-`PCX_AD_LO:5], 5'b11000 };  // 4th doubleword inside the same 256-bit data
            ahb_hwrite_o <= 0;        
	    ahb_hsize_o <= spc2ahb_size;
            ahb_hburst_o <= `AHB_INCR4;
	    ahb_hprot_o <= { spc2ahb_nc, 3'b011 };
            ahb_htrans_o <= `AHB_SEQ;
	    spc2ahb_data <= spc2ahb_packet[`PCX_DA_HI:`PCX_DA_LO];      

// Print details of return packet
`ifdef DEBUG
        $display("INFO: AHB2SPC: *** RETURN PACKET TO SPARC CORE ***");	 
        $display("INFO: AHB2SPC: Valid bit is %X", ahb2spc_valid);
        case(ahb2spc_type)
          `IFILL_RET: $display("INFO: AHB2SPC: Return Packet of Type IFILL_RET");
          `LOAD_RET: $display("INFO: AHB2SPC: Return Packet of Type LOAD_RET");
          `ST_ACK: $display("INFO: AHB2SPC: Return Packet of Type ST_ACK");
          default: $display("INFO: AHB2SPC: Return Packet of Type Unknown");
        endcase
        $display("INFO: AHB2SPC: L2 Miss is %X", ahb2spc_miss);
        $display("INFO: AHB2SPC: Error is %X", ahb2spc_error);
        $display("INFO: AHB2SPC: Non-Cacheable bit is %X", ahb2spc_nc);
        $display("INFO: AHB2SPC: Thread is %X", ahb2spc_thread);
        $display("INFO: AHB2SPC: Way Valid is %X", ahb2spc_way_valid);
        $display("INFO: AHB2SPC: Replaced L2 Way is %X", ahb2spc_way);
        $display("INFO: AHB2SPC: Fetch for Boot is %X", ahb2spc_boot_fetch);
        $display("INFO: AHB2SPC: Atomic LD/ST or 2nd IFill Packet is %X", ahb2spc_atomic);
        $display("INFO: AHB2SPC: PFL is %X", ahb2spc_pfl); 
        $display("INFO: AHB2SPC: Data is %X", ahb2spc_data);
`endif

            waited_phase <= 3'b100;
       	    burst_phase <= 3'b111;

          end else begin
            state <= `STATE_WAITED;  
            ahb_htrans_o <= `AHB_IDLE; 
          end
         // Data phase of fourth burst
        end else if(burst_phase == 3'b11) begin

          state_prev <= 3'b100;

	  // Wait until AMBA AHB access completes
          if(ahb_hresp_i==`AHB_OKAY && ahb_hready_i==1) begin

            // Latch the data and set up the return packet for the SPARC Core
            ahb2spc_atomic <= 1'b1;
            ahb2spc_data[63:0] <= ahb_hrdata_i; //ahb2spc_data <= { ahb_hrdata_i, 64'b0 }; assume un comportamento diverso
            
            state_prev <= 3'b0;

	    spc2ahb_data <= spc2ahb_packet[`PCX_DA_HI:`PCX_DA_LO]; 
        
            burst_phase <= 3'b0;
            ahb_hburst_o <= `AHB_SINGLE;
            packet_ready <= 1'b1;
      
          end else begin
            state <= `STATE_WAITED;   
            ahb_htrans_o <= `AHB_IDLE;
          end
        // Synchronize burst phase with processor
        end else if(burst_phase == 3'b110) begin

          state_prev <= 3'b101;
          burst_prev <= 1'b1;

          if(ahb_hresp_i==`AHB_OKAY && ahb_hready_i==1) begin

            spc_ready_o <= 1;
	    state_prev <= 3'b0;
            spc_packetin_o <= ahb2spc_packet;

          end else begin

            state <= `STATE_WAITED;
            ahb_htrans_o <= `AHB_IDLE;

          end
	end

	//Phase when the packet is ready to return on the processor
        if(packet_ready == 1'b1) begin

          ahb_working <= 1'b0;

          // Return the packet to the SPARC Core
          spc_ready_o <= 1'b1;
          spc_packetin_o <= ahb2spc_packet;

          // Unconditional state change
          state <= `STATE_ADDRESS_DATA;

// Print details of return packet
`ifdef DEBUG
        $display("INFO: AHB2SPC: *** RETURN PACKET TO SPARC CORE ***");	 
        $display("INFO: AHB2SPC: Valid bit is %X", ahb2spc_valid);
        case(ahb2spc_type)
          `IFILL_RET: $display("INFO: AHB2SPC: Return Packet of Type IFILL_RET");
          `LOAD_RET: $display("INFO: AHB2SPC: Return Packet of Type LOAD_RET");
          `ST_ACK: $display("INFO: AHB2SPC: Return Packet of Type ST_ACK");
          default: $display("INFO: AHB2SPC: Return Packet of Type Unknown");
        endcase
        $display("INFO: AHB2SPC: L2 Miss is %X", ahb2spc_miss);
        $display("INFO: AHB2SPC: Error is %X", ahb2spc_error);
        $display("INFO: AHB2SPC: Non-Cacheable bit is %X", ahb2spc_nc);
        $display("INFO: AHB2SPC: Thread is %X", ahb2spc_thread);
        $display("INFO: AHB2SPC: Way Valid is %X", ahb2spc_way_valid);
        $display("INFO: AHB2SPC: Replaced L2 Way is %X", ahb2spc_way);
        $display("INFO: AHB2SPC: Fetch for Boot is %X", ahb2spc_boot_fetch);
        $display("INFO: AHB2SPC: Atomic LD/ST or 2nd IFill Packet is %X", ahb2spc_atomic);
        $display("INFO: AHB2SPC: PFL is %X", ahb2spc_pfl); 
        $display("INFO: AHB2SPC: Data is %X", ahb2spc_data);
`endif

          packet_ready <= 1'b0;

        end

      // FSM State 2: STATE_WAITED
      // If the slave need to insert a wait states
      end else if(state==`STATE_WAITED) begin

        // If the bridge are processing a new request, save the data
        if(req_enable==1'b1 && ahb_working==1'b1 && read_data_saved==1'b0)begin

	  data_saved <= spc_packetout_i;
          read_data_saved <= 1'b1;

        end

        // Check if there's an incoming request
        if(spc2ahb_req==1) begin

          req_enable <= 1'b1;

          // Clear previously modified outputs
          spc_ready_o <= 1'b0;
          spc_packetin_o <= 0;

          // Latch target region and atomicity
          spc2ahb_region <= spc_req_i;
          spc2ahb_atomic <= spc_atom_i;	 

        // See if the interrupt vector has changed
        end else if(sys_interrupt_source_i!=ahb2spc_interrupt_source) begin

          // Set the flag for next cycle
          ahb2spc_interrupt_new <= 1;

          // Prepare the interrupt packet for the SPARC Core
          ahb2spc_valid <= 1;
          ahb2spc_type <= `INT_RET;
          ahb2spc_miss <= 0;
          ahb2spc_error <= 0;
          ahb2spc_nc <= 0;
          ahb2spc_thread <= 0;
          ahb2spc_way_valid <= 0;
          ahb2spc_way <= 0;
          ahb2spc_boot_fetch <= 0;
          ahb2spc_atomic <= 0;
          ahb2spc_pfl <= 0;	   

          // Stall other requests from the SPARC Core
          spc_stallreq_o <= 1;

        // Next cycle see if there's an int to be forwarded to the Core
        end else if(ahb2spc_interrupt_source!=6'b000000 && ahb2spc_interrupt_new) begin

          // Clean the flag
          ahb2spc_interrupt_new <= 0;

          // Send the interrupt packet to the Core
          spc_ready_o <= 1;
          spc_packetin_o <= ahb2spc_packet;

          // Stall other requests from the SPARC Core
          spc_stallreq_o <= 1;

          // Stay in this state
          state <= `STATE_ADDRESS_DATA;

        // Nothing to do, stay idle
        end else begin

          // Clear previously modified outputs
          spc_ready_o <= 0;
          
          if(burst_phase == 3'b0)
            spc_packetin_o <= 0;
  
          spc_stallreq_o <= 0;

        end
	
        // If hready HIGH, return in address data state
	if(ahb_hready_i==1'b1) begin

          state <= `STATE_ADDRESS_DATA;
          ahb_htrans_o <= `AHB_IDLE;

          // If previous phase was address data phase, read the data here to don't lost it
          if(state_prev == 3'b1) begin

           state_prev <= 3'b0;
           address_data_phase <= 1'b0;

            // Latch the data and set up the return packet for the SPARC Core
            ahb2spc_valid <= 1;
            case(spc2ahb_type)
              `IMISS_RQ: begin
                ahb2spc_type <= `IFILL_RET; // I-Cache Miss
                ahb2spc_atomic <= 0;
              end
              `LOAD_RQ: begin
                ahb2spc_type <= `LOAD_RET;  // Load
                ahb2spc_atomic <= spc2ahb_atomic;
              end
              `STORE_RQ: begin
                ahb2spc_type <= `ST_ACK;    // Store
                ahb2spc_atomic <= spc2ahb_atomic;
              end
            endcase
            ahb2spc_miss <= 0;
            ahb2spc_error <= 0;
            ahb2spc_nc <= spc2ahb_nc;
            ahb2spc_thread <= spc2ahb_thread;
            ahb2spc_way_valid <= 0;
            ahb2spc_way <= 0;
            if(spc2ahb_region==5'b10000) ahb2spc_boot_fetch <= 1;
	    else ahb2spc_boot_fetch <= 0;
            ahb2spc_pfl <= 0;	   
            if(spc2ahb_addr[3]==0) ahb2spc_data <= { ahb_hrdata_i, 64'b0 };
            else ahb2spc_data <= { 64'b0, ahb_hrdata_i };

            // See if other 64-bit AMBA AHB accesses are required
            if(
                // Instruction miss directed to RAM expects 256 bits
                ( (spc2ahb_type==`IMISS_RQ)&&(spc2ahb_region!=5'b10000) ) ||
                // Data access of 128 bits
                ( (spc2ahb_type==`LOAD_RQ)&&(spc2ahb_size==`PCX_SZ_16B) )
              ) begin
              
	      waited_phase <= 3'b10;
              
              // Issue a second request on the AMBA AHB bus
   	      ahb_hwdata_o <= 64'bZ;
              ahb_haddr_o <= { spc2ahb_region, 19'b0, spc2ahb_addr[`PCX_AD_HI-`PCX_AD_LO:4], 4'b1000 };  // 2nd doubleword inside the same quadword
	      ahb_hwrite_o <= 0;        
	      ahb_hsize_o <= spc2ahb_size;
	      ahb_hprot_o <= { spc2ahb_nc, 3'b011 };
	      ahb_htrans_o <= `AHB_SEQ;
	      spc2ahb_data <= spc2ahb_packet[`PCX_DA_HI:`PCX_DA_LO];
            end else begin

	      burst_phase <= 3'b0;
              packet_ready <= 1'b1;

            end

          // Address phase of third burst and data phase of second burst
	  end else if(state_prev == 3'b10 || (burst_prev==1 && waited_burst==1)) begin

            // Wait until AMBA AHB access completes
	    if(ahb_hresp_i==`AHB_OKAY && ahb_hready_i==1) begin

              state_prev <= 3'b0;
              burst_prev <= 0;
              state <= `STATE_ADDRESS_DATA;

              // Latch the data and set up the return packet for the SPARC Core
              ahb2spc_data[63:0] <= ahb_hrdata_i;
                 
              // Issue a third request on the AMBA AHB bus
	      ahb_hwdata_o <= 64'bZ;
              ahb_haddr_o <= { spc2ahb_region, 19'b0, spc2ahb_addr[`PCX_AD_HI-`PCX_AD_LO:5], 5'b10000 }; // 3nd doubleword inside the same 256-bit data
              ahb_hwrite_o <= 0;        
	      ahb_hsize_o <= spc2ahb_size;
              ahb_hburst_o <= `AHB_INCR4;
	      ahb_hprot_o <= { spc2ahb_nc, 3'b011 };
	      ahb_htrans_o <= `AHB_SEQ;
	      spc2ahb_data <= spc2ahb_packet[`PCX_DA_HI:`PCX_DA_LO]; 
 
              // See if two return packets are required or just one
              if(spc2ahb_type==`IMISS_RQ && spc2ahb_region==5'b10000)
                packet_ready <= 1'b1;
              else begin
                if(waited_burst==1) begin
                  waited_phase <= 3'b100;
                  burst_phase <= 3'b111;
                  waited_burst <= 0;
                end else begin
                  waited_phase <= 3'b11;
                  burst_phase <= 3'b110;
                  waited_burst <= 0;
                end
              end
            end

          // Address phase of fourth burst and data phase of third burst
          end else if(state_prev==3'b11) begin

            // Wait until AMBA AHB access completes
	    if(ahb_hresp_i==`AHB_OKAY && ahb_hready_i==1) begin

              state_prev <= 3'b0;
              state <= `STATE_ADDRESS_DATA;

              // Latch the data and set up the return packet for the SPARC Core
              ahb2spc_data <= { ahb_hrdata_i, 64'b0 };

              // Issue a fourth request on the AMBA AHB bus
	      ahb_hwdata_o <= 64'bZ;
              ahb_haddr_o <= { spc2ahb_region, 19'b0, spc2ahb_addr[`PCX_AD_HI-`PCX_AD_LO:5], 5'b11000 };  // 4th doubleword inside the same 256-bit data
              ahb_hwrite_o <= 0;        
	      ahb_hsize_o <= spc2ahb_size;
              ahb_hburst_o <= `AHB_INCR4;
	      ahb_hprot_o <= { spc2ahb_nc, 3'b011 };
              ahb_htrans_o <= `AHB_SEQ;
	      spc2ahb_data <= spc2ahb_packet[`PCX_DA_HI:`PCX_DA_LO];      

// Print details of return packet
`ifdef DEBUG
        $display("INFO: AHB2SPC: *** RETURN PACKET TO SPARC CORE ***");	 
        $display("INFO: AHB2SPC: Valid bit is %X", ahb2spc_valid);
        case(ahb2spc_type)
          `IFILL_RET: $display("INFO: AHB2SPC: Return Packet of Type IFILL_RET");
          `LOAD_RET: $display("INFO: AHB2SPC: Return Packet of Type LOAD_RET");
          `ST_ACK: $display("INFO: AHB2SPC: Return Packet of Type ST_ACK");
          default: $display("INFO: AHB2SPC: Return Packet of Type Unknown");
        endcase
        $display("INFO: AHB2SPC: L2 Miss is %X", ahb2spc_miss);
        $display("INFO: AHB2SPC: Error is %X", ahb2spc_error);
        $display("INFO: AHB2SPC: Non-Cacheable bit is %X", ahb2spc_nc);
        $display("INFO: AHB2SPC: Thread is %X", ahb2spc_thread);
        $display("INFO: AHB2SPC: Way Valid is %X", ahb2spc_way_valid);
        $display("INFO: AHB2SPC: Replaced L2 Way is %X", ahb2spc_way);
        $display("INFO: AHB2SPC: Fetch for Boot is %X", ahb2spc_boot_fetch);
        $display("INFO: AHB2SPC: Atomic LD/ST or 2nd IFill Packet is %X", ahb2spc_atomic);
        $display("INFO: AHB2SPC: PFL is %X", ahb2spc_pfl); 
        $display("INFO: AHB2SPC: Data is %X", ahb2spc_data);
`endif

              waited_phase <= 3'b100;
              burst_phase <= 3'b111;

            end

          // Data phase of fourth burst
          end else if(state_prev == 3'b100) begin

            // Wait until AMBA AHB access completes
            if(ahb_hresp_i==`AHB_OKAY && ahb_hready_i==1) begin

              state_prev <= 3'b0;
              state <= `STATE_ADDRESS_DATA;

              // Latch the data and set up the return packet for the SPARC Core
              ahb2spc_data[63:0] <= ahb_hrdata_i;
              burst_phase <= 3'b0;
	      ahb_hburst_o <= `AHB_SINGLE;
              packet_ready <= 1'b1;

            end 
          end
        end else begin

	  // Check if there's an incoming request
          if(spc2ahb_req==1 && req_enable==1'b0) begin

            state <= `STATE_ADDRESS_DATA;

            req_enable <= 1'b1;

            // Clear previously modified outputs
            spc_ready_o <= 1'b0;
            spc_packetin_o <= 0;

            // Latch target region and atomicity
            spc2ahb_region <= spc_req_i;
            spc2ahb_atomic <= spc_atom_i;	 

            // Stall other requests from the SPARC Core
            spc_stallreq_o <= 1;

          // See if the interrupt vector has changed
          end else if(sys_interrupt_source_i!=ahb2spc_interrupt_source) begin

            // Set the flag for next cycle
            ahb2spc_interrupt_new <= 1;

            // Prepare the interrupt packet for the SPARC Core
            ahb2spc_valid <= 1;
            ahb2spc_type <= `INT_RET;
            ahb2spc_miss <= 0;
            ahb2spc_error <= 0;
            ahb2spc_nc <= 0;
            ahb2spc_thread <= 0;
            ahb2spc_way_valid <= 0;
            ahb2spc_way <= 0;
            ahb2spc_boot_fetch <= 0;
            ahb2spc_atomic <= 0;
            ahb2spc_pfl <= 0;	   

            // Stall other requests from the SPARC Core
            spc_stallreq_o <= 1;

          // Next cycle see if there's an int to be forwarded to the Core
          end else if(ahb2spc_interrupt_source!=6'b000000 && ahb2spc_interrupt_new) begin

            // Clean the flag
            ahb2spc_interrupt_new <= 0;

            // Send the interrupt packet to the Core
            spc_ready_o <= 1;
            spc_packetin_o <= ahb2spc_packet;

            // Stall other requests from the SPARC Core
            spc_stallreq_o <= 1;
  
            // Stay in this state
            state <= `STATE_WAITED;

          // Nothing to do, stay idle
          end else begin

	    state <= `STATE_WAITED;

            // Clear previously modified outputs
            spc_ready_o <= 0;
	    if(burst_phase == 3'b0)
              spc_packetin_o <= 0;
            spc_stallreq_o <= 0;

          end
        end
        
        // Synchronize burst phase with processor
        if(state_prev==3'b101) begin
          
          if(ahb_hresp_i==`AHB_OKAY && ahb_hready_i==1)begin

            state <= `STATE_ADDRESS_DATA;
            state_prev <= 3'b0;
                 
            spc_packetin_o <= ahb2spc_packet;
            spc_ready_o <= 1;

          end
        end

         // Waited phase to synchronize processor and bridge
        if(waited_phase==3'b1) begin

          // Clear previously modified outputs
          ahb_haddr_o <= 64'b0;
          ahb_hwrite_o <= 1'b0;
          ahb_hsize_o <= 3'b0;
	  ahb_hprot_o <= 4'b0;
          ahb_htrans_o <= `AHB_IDLE;
          ahb_hmastlock_o <= 1'b0;
          ahb_hwdata_o <= 64'bZ;

	  waited_phase <= 3'b0;
          address_data_phase <= 1'b1;
	    
        end

         // Waited phase to synchronize processor and bridge
        if(waited_phase==3'b10) begin

          // Clear previously modified outputs
          ahb_haddr_o <= 64'b0;
          ahb_hwrite_o <= 1'b0;
          ahb_hsize_o <= 3'b0;
	  ahb_hprot_o <= 4'b0;
          ahb_htrans_o <= `AHB_IDLE;
          ahb_hmastlock_o <= 1'b0;
          ahb_hwdata_o <= 64'bZ;

	  waited_phase <= 3'b0;
          burst_phase <= 3'b001;
	    
        end

         // Waited phase to synchronize processor and bridge
        if(waited_phase==3'b11) begin

          // Clear previously modified outputs
          ahb_haddr_o <= 64'b0;
          ahb_hwrite_o <= 1'b0;
          ahb_hsize_o <= 3'b0;
	  ahb_hprot_o <= 4'b0;
          ahb_htrans_o <= `AHB_IDLE;
          ahb_hmastlock_o <= 1'b0;
          ahb_hwdata_o <= 64'bZ;

	  waited_phase <= 3'b0;
          burst_phase <= 3'b010;
	    
        end

         // Waited phase to synchronize processor and bridge
        if(waited_phase==3'b100) begin

          // Clear previously modified outputs
          ahb_haddr_o <= 64'b0;
          ahb_hwrite_o <= 1'b0;
          ahb_hsize_o <= 3'b0;
	  ahb_hprot_o <= 4'b0;
          ahb_htrans_o <= `AHB_IDLE;
          ahb_hmastlock_o <= 1'b0;
          ahb_hwdata_o <= 64'bZ;

	  waited_phase <= 3'b0;
          burst_phase <= 3'b011;
          waited_burst <= 0;
	    
        end

      end else begin
        
        // Clear AMBA AHB Master interface outputs
        ahb_haddr_o <= `AHB_ADDR_WIDTH'b0;
        ahb_hwrite_o <= 0;
        ahb_hsize_o <= `AHB_SIZE_WIDTH'b0;
        ahb_hburst_o <= `AHB_SINGLE;
        ahb_hprot_o <= `AHB_PROT_WIDTH'b0;
        ahb_htrans_o <= `AHB_IDLE;
        ahb_hmastlock_o <= 0;
        ahb_hwdata_o <= `AHB_DATA_WIDTH'bZ;

      end
    end
  end

endmodule
